#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 

from talos.thrift.auth.ttypes import Credential
from talos.thrift.auth import constants
from talos.thrift.transport.TTransport import TTransportBase
from talos.thrift.common.ttypes import GalaxyTalosException
from io import BytesIO
from datetime import datetime
from six.moves import urllib
from six.moves import http_client
import six
import time
import hashlib
import hmac
import base64
import os
import ssl
import sys
import uuid
import logging
import dns.resolver
import random
import traceback


class TalosHttpClientTransportFactory:

    _url = str
    _credential = Credential
    _client = http_client.HTTPConnection
    _clockOffset = int
    _agent = str

    def __init__(self, url=None, credential=None, client=None, agent=None):
        self._url = url
        self._credential = credential
        self._client = client
        self._agent = agent

    def get_transport(self, trans=None):
        return self.get_transport_with_clock_offset(trans, 0, "")

    def get_transport_with_clock_offset(self, trans=None, clockOffset=None, query=None):
        if trans:
            pass
        t2 = TalosHttpClient(uri_or_host=self._url, credential=self._credential, clockOffset=self._clockOffset, query=query)
        return t2


class TalosHttpClient(TTransportBase):
    logger = logging.getLogger("TalosHttpClient")
    GMT_FORMAT = '%a, %d %b %Y %H:%M:%S GMT'
    _REQUEST_ID_LENGTH = 8
    _url_ = None
    _requestBuffer_ = BytesIO
    _inputStream_ = BytesIO()
    _connectTimeout_ = 0
    _socketTimeout_ = 0
    _clockOffset = 0
    _queryString = None
    _supportAccountKey = False
    credential = Credential

    __http = None
    scheme = None
    using_proxy = None
    host = str
    port = None
    __timeout = None
    keyfile = None
    certfile = None
    context = None
    realhost = None
    realport = None
    proxy_auth = None
    path = None
    __http_response = None
    __wbuf = None
    __custom_headers = None
    __auth_headers = None
    code = None
    message = None
    _headers = dict()
    _uri_or_host = None
    _sid = None

    # def __init__(self, uri_or_host, credential: Credential, httpClient: http.client.HTTPConnection,
    #              agent: str, clockOffset: int, query: str):
    #     self._queryString = query
    #     self.__http = httpClient
    #     self.__init__(uri_or_host, credential, clockOffset, True)

    def __init__(self, uri_or_host, credential=None, clockOffset=None, supportAccountKey=None,
                port=None, path=None, cafile=None, cert_file=None, key_file=None, ssl_context=None):
        """THttpClient supports two different types of construction:

        THttpClient(host, port, path) - deprecated
        THttpClient(uri, [port=<n>, path=<s>, cafile=<filename>, cert_file=<filename>, key_file=<filename>, ssl_context=<context>])

        Only the second supports https.  To properly authenticate against the server,
        provide the client's identity by specifying cert_file and key_file.  To properly
        authenticate the server, specify either cafile or ssl_context with a CA defined.
        NOTE: if both cafile and ssl_context are defined, ssl_context will override cafile.
        """
        if port is not None:
            self.logger.warning(
                "Please use the THttpClient('http{s}://host:port/path') constructor",
                DeprecationWarning,
                stacklevel=2)
            self.host = uri_or_host
            self.port = port
            assert path
            self.path = path
            self.scheme = 'http'
        else:
            parsed = urllib.parse.urlparse(uri_or_host)
            self.scheme = parsed.scheme
            assert self.scheme in ('http', 'https')
            if self.scheme == 'http':
                self.port = parsed.port or http_client.HTTPConnection.default_port
            elif self.scheme == 'https':
                self.port = parsed.port or http_client.HTTPConnection.default_port
                self.certfile = cert_file
                self.keyfile = key_file
                self.context = ssl.create_default_context(cafile=cafile) if (
                            cafile and not ssl_context) else ssl_context
            self.host = parsed.hostname
            self.path = parsed.path
            if parsed.query:
                self.path += '?%s' % parsed.query
        try:
            proxy = urllib.request.getproxies()[self.scheme]
        except KeyError:
            proxy = None
        else:
            if urllib.request.proxy_bypass(self.host):
                proxy = None
        if proxy:
            parsed = urllib.parse.urlparse(proxy)
            self.realhost = self.host
            self.realport = self.port
            self.host = parsed.hostname
            self.port = parsed.port
            self.proxy_auth = self.basic_proxy_auth_header(parsed)
        else:
            self.realhost = self.realport = self.proxy_auth = None
        self._clockOffset = clockOffset
        self._supportAccountKey = supportAccountKey
        self.credential = credential
        self.__wbuf = BytesIO()
        self.__http_response = None
        self.__timeout = None
        self._uri_or_host = uri_or_host

    @staticmethod
    def basic_proxy_auth_header(proxy):
        if proxy is None or not proxy.username:
            return None
        ap = "%s:%s" % (urllib.parse.unquote(proxy.username),
                        urllib.parse.unquote(proxy.password))
        cr = base64.b64encode(ap).strip()
        return "Basic " + cr

    def using_proxy(self):
        return self.realhost is not None

    def set_timeout(self, ms):
        if ms is None:
            self.__timeout = None
        else:
            self.__timeout = int(ms / 1000.0)

    def set_custom_headers(self, headers):
        self.__custom_headers = headers

    def set_support_account_key(self, supportAccountKey=None):
        self._supportAccountKey = supportAccountKey

    # Set signature related headers when credential is properly set
    def set_authentication_headers(self, data=None):
        if self.credential:
            if self.credential.secretKey and self.credential.type:
                # signature is supported
                if constants.SIGNATURE_SUPPORT.get(self.credential.type):
                    # host
                    host = str(self.host)
                    host = host.split(":")[0]
                    self.__http.putheader(constants.HK_HOST, host)
                    self._headers[str(constants.HK_HOST).lower()] = host

                    # timestamp
                    timestamp = int(time.time()) + self._clockOffset
                    self.__http.putheader(constants.HK_TIMESTAMP, str(timestamp))
                    self._headers[str(constants.HK_TIMESTAMP).lower()] = str(timestamp)
                    mi_date = datetime.utcnow().strftime(self.GMT_FORMAT)
                    self.__http.putheader(constants.MI_DATE, mi_date)
                    self._headers[str(constants.MI_DATE).lower()] = mi_date

                    # content md5
                    m = hashlib.md5()
                    m.update(data)
                    data_md5 = m.hexdigest()
                    self.__http.putheader(constants.HK_CONTENT_MD5, data_md5)
                    self._headers[str(constants.HK_CONTENT_MD5).lower()] = data_md5

                    # signature
                    headers = dict()
                    for (k, v) in self._headers.items():
                        headers[str(k).lower()] = str(v)

                    # generate authString for header
                    try:
                        authType = "Galaxy-V2 "
                        if self._supportAccountKey:
                            authType = "Galaxy-V3 "
                        if self.credential.type == constants.UserType.APP_ACCESS_TOKEN:
                            authType = "OAuth "

                        # check secretKeyId and set header attached info
                        # secretKeyId format: "Service-Admin#SecretKeyId#developerId"
                        secretKeyId = self.credential.secretKeyId
                        if constants.HK_SERVICE_ADMIN in secretKeyId:
                            items = str(secretKeyId).split(constants.HK_SERVICE_MARK)
                            if not len(items) == 3:
                                raise RuntimeError("Invalid credential secretKeyId, " + \
                                                   "expected: 3, actual: " + str(len(items)))
                            #  reset secretKeyId and add header attached info
                            secretKeyId = items[0]
                            self.__http.putheader(constants.HK_ATTACHED_INFO, items[2])
                            self._headers[str(constants.HK_ATTACHED_INFO).lower()] = items[2]

                        authString = authType + secretKeyId + ":"
                        authString += self.generate_auth_header(self._headers, self.canonicalize_resource(self._uri_or_host))
                        self.__http.putheader(constants.HK_AUTHORIZATION, authString)
                        self._headers[str(constants.HK_AUTHORIZATION).lower()] = authString
                    except Exception as e:
                        self.logger.error("fail to sign" + str(traceback.format_exc()))
            else:
                if self.credential.type == constants.UserType.APP_XIAOMI_SSO:
                    authString = "SSO " + self._sid + ":" + self.credential.secretKey + ":" + self.credential.secretKeyId
                    self.__http.putheader(constants.HK_AUTHORIZATION, authString)
                elif self.credential.type == constants.UserType.APP_ANONYMOUS:
                    authString = "Guest " + self.credential.secretKeyId
                    self.__http.putheader(constants.HK_AUTHORIZATION, authString)
                else:
                    raise RuntimeError("Unsupported user type: " + self.credential.type)
            return

    def generate_auth_header(self, headers=None, canonicalizeResource=None):
        stringToSign = "POST\n"
        if "content-md5" in headers:
            stringToSign += headers["content-md5"]
        stringToSign += '\n'
        if "content-type" in headers:
            stringToSign += headers["content-type"]
        stringToSign += '\n'
        stringToSign += self.canonicalize_xiaomi_headers(headers) + canonicalizeResource
        key = str(self.credential.secretKey).encode("utf-8")
        msg = stringToSign.encode("utf-8")
        mac = hmac.new(key, msg, hashlib.sha1)
        returnStr = str(base64.b64encode(mac.digest()).decode("utf-8"))
        return returnStr

    def canonicalize_xiaomi_headers(self, headers=None):
        if not headers:
            return ""
        else:
            canonicalizedKeys = list()
            canonicalizedHeaders = dict()
            for (k, v) in headers.items():
                lowerKey = str(k).lower()
                if "x-xiaomi-" in lowerKey:
                    canonicalizedKeys.append(lowerKey)
                    canonicalizedHeaders[lowerKey] = v + '\n'
            canonicalizedKeys.sort()
            result = "\n"
            for k in canonicalizedKeys:
                result += k
                result += ':'
                result += canonicalizedHeaders[k]
            return result

    def canonicalize_resource(self, uri=None):
        subResource = ["acl", "quota", "uploads", "partNumber", "uploadId", "storageAccessToken", "metadata"]
        parsedUrl = http_client.urlsplit(url=uri)
        result = parsedUrl.path
        queryArgs = parsedUrl.query
        canonicalizeQuery = []
        for k in queryArgs:
            if k in subResource:
                canonicalizeQuery.append(k)
        if not len(canonicalizeQuery) == 0:
            i = 0
            canonicalizeQuery.sort()
            for k in canonicalizeQuery:
                if i == 0:
                    result += '?'
                else:
                    result += '&'
                values = queryArgs[k]
                if len(values) == 1 and values[0] == "":
                    result += k
                else:
                    result += k
                    result += '='
                    result += values[len(values) - 1]
                i += 1

        return result

    def isOpen(self):
        return self.__http is not None

    def open(self):
        if self.scheme == 'http':
            self.__http = http_client.HTTPConnection(self.get_addr_by_host(self.host), self.port,
                                                     timeout=self.__timeout)
        elif self.scheme == 'https':
            self.__http = http_client.HTTPSConnection(self.get_addr_by_host(self.host), self.port,
                                                      key_file=self.keyfile,
                                                      cert_file=self.certfile,
                                                      timeout=self.__timeout,
                                                      context=self.context)
        if self.using_proxy():
            self.__http.set_tunnel(self.get_addr_by_host(self.realhost), self.realport,
                                   {"Proxy-Authorization": self.proxy_auth})

    def close(self):
        self.__http.close()
        self.__http = None
        self.__http_response = None

    def read(self, sz):
        return self.__http_response.read(sz)

    def write(self, buf):
        self.__wbuf.write(buf)

    def flush(self):
        if self.isOpen():
            self.close()
        self.open()

        # Pull data out of buffer
        data = self.__wbuf.getvalue()
        self.__wbuf = BytesIO()

        # HTTP request
        requestId = self.generate_random_id(self._REQUEST_ID_LENGTH)
        sb = str(self.path)
        sb += "?id="
        sb += requestId
        if self._queryString:
            sb += "&"
            sb += self._queryString

        if self.using_proxy() and self.scheme == "http":
            # need full URL of real host for HTTP proxy here (HTTPS uses CONNECT tunnel)
            self.__http.putrequest('POST', "http://%s:%s%s" %
                                   (self.realhost, self.realport, sb))
        else:
            self.__http.putrequest('POST', sb)

        # Write headers
        self.__http.putheader('Content-Type', 'application/x-thrift-compact')
        self._headers['Content-Type'.lower()] = 'application/x-thrift-compact'
        self.__http.putheader('Content-Length', str(len(data)))
        self._headers['Content-Length'] = str(len(data))
        self.__http.putheader('Accept', 'application/x-thrift-compact')
        self._headers['Accept'.lower()] = 'application/x-thrift-compact'
        if self.using_proxy() and self.scheme == "http" and self.proxy_auth is not None:
            self.__http.putheader("Proxy-Authorization", self.proxy_auth)
            self._headers["Proxy-Authorization".lower()] = self.proxy_auth

        if not self.__custom_headers or 'User-Agent' not in self.__custom_headers:
            user_agent = 'Python/THttpClient'
            script = os.path.basename(sys.argv[0])
            if script:
                user_agent = '%s (%s)' % (user_agent, urllib.parse.quote(script))
            self.__http.putheader('User-Agent', user_agent)
            self._headers['User-Agent'.lower()] = user_agent

        if self.__custom_headers:
            for key, val in six.iteritems(self.__custom_headers):
                self.__http.putheader(key, val)
                self._headers[key] = val

        self.set_authentication_headers(data)

        self.__http.endheaders()

        # Write payload
        self.__http.send(data)

        # Get reply to flush the request
        self.__http_response = self.__http.getresponse()
        response = self.__http_response
        self.code = self.__http_response.status
        self.message = self.__http_response.reason
        self._headers = self.__http_response.msg
        if not self.code == 200:
            raise GalaxyTalosException(errorCode=self.code, errMsg=self.message)
            # pass
        # TODO Trans Exception to TalosTransportError
        # reason = response.fp.read()
        # raise TalosTransportError(self.code, self.message, current_time_mills())

    def generate_random_id(self, length=None):
        return str(uuid.uuid4())[0:length]

    def set_query_string(self, queryString=None):
        self._queryString = queryString

    def set_clock_offset(self, clockOffset=None):
        self._clockOffset = clockOffset

    def set_sid(self, sid=None):
        self._sid = sid

    # for dns cache
    def get_addr_by_host(self, host=None):
        ipList = []
        try:
            A = dns.resolver.query(host)
            if len(A.response.answer) > 0:
                for items in A.response.answer:
                    for item in items:
                        ipList.append(item)
                return str(random.choice(ipList))
        except Exception as e:
            self.logger.info("dns resolve error for host: "
                             + host + str(traceback.format_exc()))
        return host




#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 

import time
import logging
import threading
from atomic import AtomicLong
from threading import Timer
from talos.utils import Utils
from talos.utils.Utils import ReadWriteLock
from talos.utils.Utils import synchronized
from talos.thrift.message.MessageService import GetScheduleInfoRequest
import traceback


class ScheduleInfoCache:
	logger = logging.getLogger("ScheduleInfoCache")
	scheduleInfoCacheMap = dict()
	scheduleInfoMap = dict()

	def __init__(self, topicTalosResourceName=None, talosClientConfig=None, messageClient=None,
				 talosClientFactory=None):
		self.readWriteLock = ReadWriteLock()
		self.messageClientMap = dict()
		self.clientNum = AtomicLong(0)
		self.get_schedule_info_cacheLock = threading.Lock()
		if talosClientConfig:
			self.isAutoLocation = talosClientConfig.get_is_auto_location()
			self.topicTalosResourceName = topicTalosResourceName
			self.talosClientConfig = talosClientConfig
			self.messageClient = messageClient
			self.talosClientFactory = talosClientFactory

			# GetScheduleInfoScheduleExecutor
			# for Schedule get work, cause ScheduledExecutorService
			# use DelayedWorkQueue storage its task, which is unbounded.To private OOM, use
			# GetScheduleInfoExecutor execute task when transfered, setting Queue size as 2.
			self.GetScheduleInfoScheduleExecutor = Timer

			if self.isAutoLocation:
				self.logger.info("Auto location is enabled for request of " +
								 str(self.topicTalosResourceName.topicTalosResourceName))
			else:
				self.logger.info("Auto location is forbidden for request of " +
								 str(self.topicTalosResourceName.topicTalosResourceName))

			try:
				# get and update scheduleInfoMap
				self.get_schedule_info(self.topicTalosResourceName)
			except Exception as e:
				self.logger.error("Exception in GetScheduleInfoTask: ", str(traceback.format_exc()))
				if Utils.is_topic_not_exist(e):
					return

	def get_schedule_info_task(self):
		maxRetry = self.talosClientConfig.get_schedule_info_max_retry() + 1

		while maxRetry > 0:
			try:
				# get and update scheduleInfoMap
				self.get_schedule_info(self.topicTalosResourceName)
				# to prevent frequent ScheduleInfo call
				time.sleep(10)
				return
			except Exception as e:
				# 1. if throwable instance of TopicNotExist, cancel all reading task
				# 2. if other throwable such as LockNotExist or HbaseOperationFailed, retry again
				# 3. if scheduleInfoMap didn't update success after maxRetry, just return and use
				# old data, it may update next time or targeted update when Transfered.
				if Utils.is_topic_not_exist(e):
					return
				if self.logger.isEnabledFor(logging.DEBUG):
					self.logger.debug("Exception in GetScheduleInfoTask: " + str(
						traceback.format_exc()))
			finally:
				maxRetry = maxRetry - 1

	def get_schedule_info_cache(self, topicTalosResourceName=None, talosClientConfig=None,
							   messageClient=None, talosClientFactory=None):
		self.get_schedule_info_cacheLock.acquire()
		try:
			if not self.scheduleInfoCacheMap.get(topicTalosResourceName):
				if not talosClientFactory:
					# this case should not exist normally, only when interface of simpleAPI improper used
					self.scheduleInfoCacheMap[topicTalosResourceName] = ScheduleInfoCache(topicTalosResourceName,
							  talosClientConfig, messageClient)
				else:
					self.scheduleInfoCacheMap[topicTalosResourceName] = ScheduleInfoCache(topicTalosResourceName,
					talosClientConfig, messageClient, talosClientFactory)

			clientNum = self.scheduleInfoCacheMap[topicTalosResourceName].clientNum.value
			self.scheduleInfoCacheMap[topicTalosResourceName].clientNum.get_and_set(clientNum + 1)
			self.logger.info("there were " + str(self.scheduleInfoCacheMap[topicTalosResourceName].clientNum.value) +
							 " partitions of " + str(topicTalosResourceName.topicTalosResourceName)
							 + " use same scheduleInfoCache together.")
			return self.scheduleInfoCacheMap[topicTalosResourceName]
		except Exception as e:
			self.logger.error("get_schedule_info_cache func error: " + traceback.format_exc())
		finally:
			self.get_schedule_info_cacheLock.release()

	def get_or_create_message_client(self, topicAndPartition):
		if not self.scheduleInfoMap:
			self.update_schedule_info_cache()
			return self.messageClient
		host = self.scheduleInfoMap[topicAndPartition]
		if not host:
			self.update_schedule_info_cache()
			return self.messageClient
		messageClient = None
		if host in self.messageClientMap:
			messageClient = self.messageClientMap[host]
		if not messageClient:
			messageClient = self.talosClientFactory.new_message_client("http://" + str(host))
			self.messageClientMap[host] = messageClient
		return messageClient

	def update_schedule_info_cache(self):
		if self.isAutoLocation:
			self.GetScheduleInfoScheduleExecutor = Timer(interval=0, function=self.get_schedule_info_task)
			self.GetScheduleInfoScheduleExecutor.setName(
				"talos-ScheduleInfoCache-" + str(self.topicTalosResourceName.topicTalosResourceName))
			self.GetScheduleInfoScheduleExecutor.start()

	def get_is_auto_location(self):
		return self.isAutoLocation

	@synchronized
	def shut_down(self, topicTalosResourceName=None):
		if not self.scheduleInfoCacheMap[topicTalosResourceName]:
			self.logger.error("there were no scheduleInfoCache for " +
							  str(self.topicTalosResourceName.topicTalosResourceName)
							  + " when call scheduleInfoCache shutDown")
			return
		# shutdown the scheduleInfoCache when there were no client of this topic.
		if self.scheduleInfoCacheMap[topicTalosResourceName].clientNum.value == 1:
			self.logger.info("scheduleInfoCache of " + str(self.topicTalosResourceName.topicTalosResourceName)
							 + " is shutting down...")
			self.scheduleInfoCacheMap[topicTalosResourceName].GetScheduleInfoScheduleExecutor.cancel()
			if topicTalosResourceName in self.scheduleInfoCacheMap:
				del self.scheduleInfoCacheMap[topicTalosResourceName]
			self.logger.info("scheduleInfoCache of " + topicTalosResourceName + " shutdown.")
		else:
			self.logger.info("there were still " + str(self.scheduleInfoCacheMap[topicTalosResourceName].clientNum.get())
							 +" partitions of " + str(self.topicTalosResourceName.topicTalosResourceName)
							 + " use this scheduleInfoCache. skip it")

	def shut_down_all(self):
		self.logger.info("scheduleInfoCache is shutting down...")
		for key in self.scheduleInfoCacheMap.keys():
			self.scheduleInfoCacheMap[key].GetScheduleInfoScheduleExecutor.cancel()
		self.scheduleInfoCacheMap.clear()
		self.logger.info("scheduleInfoCache shutdown.")

	def get_schedule_info(self, topicTalosResourceName):
		# judge isAutoLocation serveral place to make sure request server only when need.
		# 1.before send Executor task make sure send Executor task when need;
		# 2.judge in getScheduleInfo is the Final guarantee good for code extendibility;
		if self.isAutoLocation:
			topicScheduleInfoMap = self.messageClient.get_schedule_info(
				GetScheduleInfoRequest(topicTalosResourceName))
			self.readWriteLock.write_acquire()
			self.scheduleInfoMap = topicScheduleInfoMap
			self.readWriteLock.write_release()
		if self.logger.isEnabledFor(logging.DEBUG):
			self.logger.debug("getScheduleInfo success" + str(self.scheduleInfoMap))

	# update message client just for Talos Canary
	def set_message_client(self, messageClient=None):
		self.messageClient = messageClient




#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 

from talos.client.TalosClientConfig import TalosClientConfig
from talos.thrift.auth.ttypes import Credential
from talos.thrift.common.ttypes import Version
from talos.thrift.topic import TopicService
from talos.thrift.message import MessageService
from talos.thrift.consumer import ConsumerService
from talos.thrift.quota import QuotaService
from talos.client.TalosErrors import InvalidArgumentError
from talos.client.Constants import Constants
from talos.thrift.protocol.TCompactProtocol import TCompactProtocol
from talos.client.TalosHttpClient import TalosHttpClient
import threading
import logging
import platform
import http.client
import http


class ConsumerClient:
    talosHttpClient = TalosHttpClient
    consumerClientLock = threading.Lock()

    def __init__(self, talosHttpClient=None):
        self.talosHttpClient = talosHttpClient

    def renew(self, request=None):
        self.consumerClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=renew")
            iprot = TCompactProtocol(self.talosHttpClient)
            return ConsumerService.Client(iprot).renew(request)
        except:
            pass
        finally:
            self.consumerClientLock.release()

    def lock_worker(self, request=None):
        self.consumerClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=lockWorker")
            iprot = TCompactProtocol(self.talosHttpClient)
            return ConsumerService.Client(iprot).lockWorker(request)
        except:
            pass
        finally:
            self.consumerClientLock.release()

    def query_worker(self, request=None):
        self.consumerClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=queryWorker")
            iprot = TCompactProtocol(self.talosHttpClient)
            return ConsumerService.Client(iprot).queryWorker(request)
        except:
            pass
        finally:
            self.consumerClientLock.release()

    def lock_partition(self, request=None):
        self.consumerClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=lockPartition")
            iprot = TCompactProtocol(self.talosHttpClient)
            return ConsumerService.Client(iprot).lockPartition(request)
        except:
            pass
        finally:
            self.consumerClientLock.release()

    def unlock_partition(self, request=None):
        self.consumerClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=unlockPartition")
            iprot = TCompactProtocol(self.talosHttpClient)
            return ConsumerService.Client(iprot).unlockPartition(request)
        except:
            pass
        finally:
            self.consumerClientLock.release()

    def query_offset(self, request=None):
        self.consumerClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=queryOffset")
            iprot = TCompactProtocol(self.talosHttpClient)
            return ConsumerService.Client(iprot).queryOffset(request)
        except:
            pass
        finally:
            self.consumerClientLock.release()

    def update_offset(self, request=None):
        self.consumerClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=updateOffset")
            iprot = TCompactProtocol(self.talosHttpClient)
            return ConsumerService.Client(iprot).updateOffset(request)
        except:
            pass
        finally:
            self.consumerClientLock.release()

    def get_worker_id(self, request=None):
        self.consumerClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=getWorkerId")
            iprot = TCompactProtocol(self.talosHttpClient)
            return ConsumerService.Client(iprot).getWorkerId(request).workerId
        except:
            pass
        finally:
            self.consumerClientLock.release()


class TopicClient:
    talosHttpClient = TalosHttpClient
    topicClientLock = threading.Lock()

    def __init__(self, talosHttpClient=None):
        self.talosHttpClient = talosHttpClient

    def create_topic(self, request=None):
        self.topicClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=createTopic")
            iprot = TCompactProtocol(self.talosHttpClient)
            return TopicService.Client(iprot).createTopic(request)
        except:
            pass
        finally:
            self.topicClientLock.release()

    def get_describe_info(self, request=None):
        self.topicClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=getDescribeInfo")
            iprot = TCompactProtocol(self.talosHttpClient)
            return TopicService.Client(iprot).getDescribeInfo(request)
        except:
            pass
        finally:
            self.topicClientLock.release()

    def get_topic_attribute(self, request=None):
        self.topicClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=getTopicAttribute")
            iprot = TCompactProtocol(self.talosHttpClient)
            return TopicService.Client(iprot).getTopicAttribute(request)
        except:
            pass
        finally:
            self.topicClientLock.release()

    def delete_topic(self, request=None):
        self.topicClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=deleteTopic")
            iprot = TCompactProtocol(self.talosHttpClient)
            return TopicService.Client(iprot).deleteTopic(request)
        except:
            pass
        finally:
            self.topicClientLock.release()

    def change_topic_attribute(self, request=None):
        self.topicClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=changeTopicAttribute")
            iprot = TCompactProtocol(self.talosHttpClient)
            return TopicService.Client(iprot).changeTopicAttribute(request)
        except:
            pass
        finally:
            self.topicClientLock.release()

    def list_topics(self):
        self.topicClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=listTopics")
            iprot = TCompactProtocol(self.talosHttpClient)
            return TopicService.Client(iprot).listTopics().topicInfos
        except:
            pass
        finally:
            self.topicClientLock.release()

    def set_permission(self, request=None):
        assert request.permission > 0
        self.topicClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=setPermission")
            iprot = TCompactProtocol(self.talosHttpClient)
            TopicService.Client(iprot).setPermission(request)
        except:
            pass
        finally:
            self.topicClientLock.release()

    def revoke_permission(self, request=None):
        self.topicClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=revokePermission")
            iprot = TCompactProtocol(self.talosHttpClient)
            TopicService.Client(iprot).setPermission(request)
        except:
            pass
        finally:
            self.topicClientLock.release()

    def list_permission(self, request=None):
        self.topicClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=listPermission")
            iprot = TCompactProtocol(self.talosHttpClient)
            return TopicService.Client(iprot).listPermission(request).permissions
        except:
            pass
        finally:
            self.topicClientLock.release()


class MessageClient:
    talosHttpClient = TalosHttpClient
    messageClientLock = threading.Lock()

    def __init__(self, talosHttpClient=None):
        self.talosHttpClient = talosHttpClient

    def get_message(self, request=None):
        self.messageClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=getMessage")
            iprot = TCompactProtocol(self.talosHttpClient)
            return MessageService.Client(iprot).getMessage(request)
        except:
            pass
        finally:
            self.messageClientLock.release()

    def put_message(self, request=None):
        self.messageClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=putMessage")
            iprot = TCompactProtocol(self.talosHttpClient)
            return MessageService.Client(iprot).putMessage(request)
        except:
            pass
        finally:
            self.messageClientLock.release()

    def get_topic_offset(self, request=None):
        self.messageClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=getTopicOffset")
            iprot = TCompactProtocol(self.talosHttpClient)
            return MessageService.Client(iprot).getTopicOffset(request).offsetInfoList
        except:
            pass
        finally:
            self.messageClientLock.release()

    def get_partition_offset(self, request=None):
        self.messageClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=getPartitionOffset")
            iprot = TCompactProtocol(self.talosHttpClient)
            return MessageService.Client(iprot).getPartitionOffset(request).offsetInfo
        except:
            pass
        finally:
            self.messageClientLock.release()

    def get_schedule_info(self, request=None):
        self.messageClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=getScheduleInfo")
            iprot = TCompactProtocol(self.talosHttpClient)
            return MessageService.Client(iprot).getScheduleInfo(request).scheduleInfo
        except:
            pass
        finally:
            self.messageClientLock.release()


class QuotaClient:
    talosHttpClient = TalosHttpClient
    quotaClientLock = threading.Lock()

    def __init__(self, talosHttpClient=None):
        self.talosHttpClient = talosHttpClient

    def apply_quota(self, request=None):
        self.quotaClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=applyQuota")
            iprot = TCompactProtocol(self.talosHttpClient)
            QuotaService.Client(iprot).applyQuota(request)
        except:
            pass
        finally:
            self.quotaClientLock.release()

    def revoke_quota(self, request=None):
        self.quotaClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=revokeQuota")
            iprot = TCompactProtocol(self.talosHttpClient)
            return QuotaService.Client(iprot).revokeQuota(request)
        except:
            pass
        finally:
            self.quotaClientLock.release()

    def list_quota(self):
        self.quotaClientLock.acquire()
        try:
            self.talosHttpClient.set_query_string("type=listQuota")
            iprot = TCompactProtocol(self.talosHttpClient)
            return QuotaService.Client(iprot).listQuota()
        except:
            pass
        finally:
            self.quotaClientLock.release()


class TalosClientFactory:
    logger = logging.getLogger("TalosClientFactory")
    _USER_AGENT_HEADER = "User-Agent"
    _SID = "galaxytalos"
    _DEFAULT_CLIENT_CONN_TIMEOUT = 5000

    _version = Version
    _talosClientConfig = TalosClientConfig
    _credential = Credential
    _customHeaders = dict
    _httpClient = http.client.HTTPConnection
    _agent = str
    _clockOffset = int

    def __init__(self, clientConfig=None, credential=None):
        self._talosClientConfig = clientConfig
        self._credential = credential
        self._customHeaders = None
        self._version = Version()
        self._clockOffset = 0

    def new_topic_client(self):
        headers = dict()
        headers[self._USER_AGENT_HEADER] = self.create_user_agent_header()
        if self._customHeaders:
            for k in self._customHeaders:
                headers[k] = self._customHeaders[k]
        # setting 'supportAccountKey' to true for using Galaxy-V3 auth
        talosHttpClient = TalosHttpClient(self._talosClientConfig.get_service_endpoint()
                                          + Constants.TALOS_TOPIC_SERVICE_PATH,
                                          self._credential, self._clockOffset, True)
        talosHttpClient.set_custom_headers(headers)
        talosHttpClient.set_timeout(self._talosClientConfig.get_client_conn_timeout())
        return TopicClient(talosHttpClient)

    def new_consumer_client(self):
        headers = dict()
        headers[self._USER_AGENT_HEADER] = self.create_user_agent_header()
        if self._customHeaders:
            for k in self._customHeaders:
                headers[k] = self._customHeaders[k]
        # setting 'supportAccountKey' to true for using Galaxy-V3 auth
        talosHttpClient = TalosHttpClient(self._talosClientConfig.get_service_endpoint()
                                          + Constants.TALOS_CONSUMER_SERVICE_PATH,
                                          self._credential, self._clockOffset, True)
        talosHttpClient.set_custom_headers(headers)
        talosHttpClient.set_timeout(self._talosClientConfig.get_client_conn_timeout())
        return ConsumerClient(talosHttpClient)

    def set_custom_headers(self, customHeaders=None):
        self._customHeaders = customHeaders

    def new_message_client(self, endpoint=None):
        headers = dict()
        headers[self._USER_AGENT_HEADER] = self.create_user_agent_header()
        if self._customHeaders:
            for k in self._customHeaders:
                headers[k] = self._customHeaders[k]
        # setting 'supportAccountKey' to true for using Galaxy-V3 auth
        if not endpoint:
            talosHttpClient = TalosHttpClient(self._talosClientConfig.get_service_endpoint()
                                              + Constants.TALOS_MESSAGE_SERVICE_PATH,
                                              self._credential, self._clockOffset, True)
        else:
            talosHttpClient = TalosHttpClient(endpoint + Constants.TALOS_MESSAGE_SERVICE_PATH,
                                              self._credential, self._clockOffset, True)
        talosHttpClient.set_custom_headers(headers)
        talosHttpClient.set_timeout(self._talosClientConfig.get_client_conn_timeout())
        return MessageClient(talosHttpClient)

    def check_credential(self):
        if not self._credential:
            raise InvalidArgumentError("Credential is not set")

    def create_user_agent_header(self):
        return "Python-SDK/" + str(self._version.major) + "." + str(self._version.minor) + "." + str(self._version.revision) + " Python/" + platform.python_version()


